#include <stdio.h>
#include "geocalc.h"
#include "charimg.h"

void demo_simple();					//演示基本的 ECEF 和 LLA 互换
void demo_perform();				//演示ECEF和LLA互换的性能
void demo_ellips_geodesic();		//演示由AB两点确定的测地线
void demo_ellips_range();			//演示由 A 和初始方向、距离计算的测地线
void demo_observe_azel();			//演示求取观测方位俯仰角
void demo_observe_vis();			//演示判断两个位置会不会被地表遮挡
void demo_cover_grond_driven();		//演示由地表的位置发起计算，查阅场强, 绘制点阵图
void demo_cover_slide_driven();		//演示由幻灯（或者天线）发起计算，获得地表的交点，绘制矢量图。
void demo_contour_distance();		//演示绘制距离等值线。
void demo_contour_elevation();		//演示绘制仰角等值线。

int main(int /*argc*/, char * /*argv*/[])
{
	demo_simple();
	demo_perform();
	demo_ellips_geodesic();
	demo_ellips_range();
	demo_observe_azel();
	demo_observe_vis();
	demo_cover_grond_driven();
	demo_cover_slide_driven();
	demo_contour_distance();
	demo_contour_elevation();
	return 0;
}

//!
//! \brief demo_simple 演示基本的 ECEF 和 LLA 互换
//!
void demo_simple()
{
	printf ("\n\n========%s=======\n",__FUNCTION__);
	const double LLA[] = {35.13635472, 107.37464, 3548.772};
	double ECEF[] = {0, 0, 0};

	using namespace CES_GEOCALC;

	printf("LAT=%.7lf\nLON=%.7lf\nALT=%.7lf\n\n", LLA[0], LLA[1], LLA[2]);

	lla2ecef(LLA, ECEF);

	printf("X  =%.7lf\nY  =%.7lf\nZ  =%.7lf\n\n", ECEF[0], ECEF[1], ECEF[2]);

	double LLA2[] = {0, 0, 0};
	ecef2lla(ECEF,LLA2);

	printf("LAT=%.7lf\nLON=%.7lf\nALT=%.7lf\n\n", LLA2[0], LLA2[1], LLA2[2]);

}

//!
//! \brief demo_perform 演示ECEF和LLA互换的性能
//!
void demo_perform()
{
	printf ("\n\n========%s=======\n",__FUNCTION__);
	using namespace CES_GEOCALC;
	for (double lon = -180; lon <=180; lon+=60)
	{
		for (double lat = -90; lat <=90; lat += 15)
		{
			double LLA[] = {lat,lon,10000};
			double ECEF[] = {0, 0, 0};
			double LLA2[] = {0, 0, 0};
			lla2ecef(LLA, ECEF);
			int iter = 0;
			ecef2lla(ECEF,LLA2,nullptr,nullptr,32,&iter);
			printf("LAT=%12.7lf,LON=%12.7lf,dLAT=%10.7lf,dLON=%10.7lf,dALT=%5.3lf, iter = %d\n",
				   lat,lon,
				   LLA[0]-LLA2[0], LLA[1]-LLA2[1], LLA[2]-LLA2[2],iter);
		}
	}
}

//!
//! \brief demo_ellips_geodesic  演示由AB两点确定的测地线
//!
void demo_ellips_geodesic()
{
	printf ("\n\n========%s=======\n",__FUNCTION__);
	printf("\n\n");
	const double LLA_A[] = {31.847774515, 117.292115092, 10000};
	const double LLA_B[] = {37.807300349, -122.366731167,10000};

	const int maxRes = 32768;
	double (* lla)[3] = new double [maxRes][3];
	double (* ecef)[3] = new double [maxRes][3];
	double * distance = new double [maxRes];
	double * direction = new double [maxRes];
	double * err = new double [maxRes];
	using namespace CES_GEOCALC;

	const int step = 1000;

	int res = ellips_geodesic(LLA_A,LLA_B,maxRes,100000/step,lla,ecef,distance,direction,err,false,true,step);

	for(int i=0;i<res;++i)
	{
		printf("I=%d LAT=%12.7lf LON=%12.7lf ALT=%7.1lf DIS=%8.1lf DIR=%12.7lf ERR=%lf\n",
			   i,
			   lla[i][0],lla[i][1],lla[i][2],
			   distance[i],direction[i],err[i]
			   );
	}

	delete [] lla;
	delete [] ecef;
	delete [] distance;
	delete [] direction;
	delete [] err;
}

//!
//! \brief demo_ellips_range 演示由 A 和初始方向计算的测地线
//!
void demo_ellips_range()
{
	printf ("\n\n========%s=======\n",__FUNCTION__);
	const double LLA_A[] = {31.847774515, 117.292115092, 10000};
	const double range = 12345000;
	const double C = 45;
	const int maxRes = 32768;
	double (* lla)[3] = new double [maxRes][3];
	double (* ecef)[3] = new double [maxRes][3];
	double * distance = new double [maxRes];
	double * direction = new double [maxRes];
	double * err = new double [maxRes];
	using namespace CES_GEOCALC;

	const int step = 1000;

	int res = ellips_range(LLA_A,range,C,maxRes,100000/step,lla,ecef,distance,direction,err,false,true,step);

	for(int i=0;i<res;++i)
	{
		printf("I=%d LAT=%12.7lf LON=%12.7lf ALT=%7.1lf DIS=%8.1lf DIR=%12.7lf ERR=%lf\n",
			   i,
			   lla[i][0],lla[i][1],lla[i][2],
			   distance[i],direction[i],err[i]
			   );
	}

	delete [] lla;
	delete [] ecef;
	delete [] distance;
	delete [] direction;
	delete [] err;
}

//!
//! \brief demo_observe_azel	演示求取方位俯仰角
//!
void demo_observe_azel()
{
	printf ("\n\n========%s=======\n",__FUNCTION__);
	const double LLA_A[] = {31.847774515, 117.292115092, 10000};
	const double LLA_B[] = {-2.27364, 56.366731167,356700234};
	double el = 0, az = 0, range = 0;
	using namespace CES_GEOCALC;

	observe_az_el_range(LLA_A,LLA_B,&az,&el,&range,true);

	printf("Az = %lf, El=%lf, Range=%lf meters.\n", az,el,range);

}

//!
//! \brief demo_observe_vis	演示判断两个位置会不会被遮挡
//!
void demo_observe_vis()
{
	printf ("\n\n========%s=======\n",__FUNCTION__);
	const double LLA_A[] = {31.84, 117.292, 2117};
	const double LLA_B[] = {30.27, 114.344, 8092};
	const double LLA_C[] = {31.83, 117.273, 9092};
	using namespace CES_GEOCALC;

	bool bViz= observe_vis_judge_el(LLA_A,LLA_B);

	printf("VIS=%d.\n", bViz?1:0);

	bViz= observe_vis_judge_el(LLA_A,LLA_B,5);

	printf("VIS=%d.\n", bViz?1:0);

	bViz= observe_vis_judge_el(LLA_A,LLA_C,0);

	printf("VIS=%d.\n", bViz?1:0);

}

//!由地表的位置发起计算，查阅强度, 绘制点阵图
void demo_cover_grond_driven()
{
	printf ("\n\n========%s=======\n",__FUNCTION__);
	using namespace CES_GEOCALC;
	const double earth_H = 100;//地表海拔
	//光源的LLA
	const double lla_torch [3] = {35.273, 117.121, 10000};
	//光源的姿态角度
	const double hpr_torch [3] = {135,-50,150};
	double T_torch[3][3] = {0,0,0,0,0,0,0,0,0};
	hpr_T_in_enu(hpr_torch,T_torch);
	//光源 ECEF
	double ecef_torch[3] = {0,0,0};
	lla2ecef(lla_torch,ecef_torch);
	//光源ENU
	double R_torch[3][3] = {0,0,0,0,0,0,0,0,0};
	enu_R_from_lla(lla_torch,R_torch);
	//光源膜片（强度图）
	unsigned char cimg[64][64];
	cimg_matrix(cimg);
	//计算中心交点。手电的指向就是H轴在ECEF中的指向kH_ecef。
	const double kH_enu[3] = {T_torch[0][0],T_torch[0][1],T_torch[0][2]};
	double kH_ecef[3] = {0,0,0};
	enu2ecef(R_torch,kH_enu,kH_ecef);

	double intersecs_ecef[2][3], intersecs_t[2];
	int c_inter = ellipsoid_intersections(ecef_torch,kH_ecef,earth_H,intersecs_ecef,intersecs_t);
	double lla_i[3] = {0,0,0};
	if (c_inter)
		ecef2lla(intersecs_ecef[0],lla_i);

	printf ("\nIntersec: %lf,%lf,%lf\n",lla_i[0],lla_i[1],lla_i[2]);
	//穷举测量点M
	int output = 0;
	for (double dlat=lla_i[0]-0.5;dlat<lla_i[0]+0.5;dlat+=0.001)
	{
		for (double dlon=lla_i[1]-0.5;dlon<lla_i[1]+0.5;dlon+=0.001)
		{
			//M 的LLA、ECEF
			const double lla_M [3] = {dlat, dlon, earth_H};
			//对地表的位置，直接通过仰角>0判断覆盖
			double el = 0;
			observe_az_el_range(lla_M,lla_torch,0,&el,0);
			if (el<0)
				continue;
			double ecef_M[3] = {0,0,0};
			lla2ecef(lla_M,ecef_M);
			//M 在光源ENU中的坐标
			double enu_M[3]={0,0,0};
			ecef2enu(R_torch,ecef_torch,ecef_M,enu_M);

			//M 在光源 HPR中的坐标
			double hpr_M[3] = {0,0,0};
			enu2HPR(T_torch,enu_M,hpr_M);

			//alpha beta D
			const double H = hpr_M[0], P=hpr_M[1], R=hpr_M[2];
			const double D = sqrt(H*H+P*P+R*R);
			const double alpha = rad2deg * acos(H/D);
			const double beta = rad2deg * ((P>=-1e-10 && P<1e-10 && R>=-1e-10 && R<1e-10 )?0:-atan2(R,P));

			const unsigned char A = F_cimg(alpha,beta,cimg);
			if (A && (output++ % 100==0) )
				printf ("ID %d lat=%.6lf,lon=%.6lf,D=%lf,Alpha=%lf,Beta=%lf,A=%d\n",output,lla_M[0],lla_M[1],D,alpha,beta,(int)A);
			//printf ("%.12lf,%.12lf\n",lla_M[0],lla_M[1]);
		}
	}
}

//由幻灯片（或者天线）发起计算，获得地表的交点，绘制矢量图。
void demo_cover_slide_driven()
{
	printf ("\n\n========%s=======\n",__FUNCTION__);
	using namespace CES_GEOCALC;
	//光源的LLA
	const double lla_torch [3] = {35.273, 117.121, 10000};
	//光源的姿态角度
	const double hpr_torch [3] = {135,-50,150};
	double T_torch[3][3] = {0,0,0,0,0,0,0,0,0};
	hpr_T_in_enu(hpr_torch,T_torch);
	//光源 ECEF
	double ecef_torch[3] = {0,0,0};
	lla2ecef(lla_torch,ecef_torch);
	//光源ENU
	double R_torch[3][3] = {0,0,0,0,0,0,0,0,0};
	enu_R_from_lla(lla_torch,R_torch);

	//计算中心交点。手电的指向就是H轴在ECEF中的指向kH_ecef。
	const double kH_enu[3] = {T_torch[0][0],T_torch[0][1],T_torch[0][2]};
	double kH_ecef[3] = {0,0,0};
	enu2ecef(R_torch,kH_enu,kH_ecef);

	double intersecs_ecef[2][3], intersecs_t[2];
	const double earth_H = 100;//地表海拔
	int c_inter = ellipsoid_intersections(ecef_torch,kH_ecef,earth_H,intersecs_ecef,intersecs_t);
	double lla_i[3] = {0,0,0};
	if (c_inter)
		ecef2lla(intersecs_ecef[0],lla_i);

	printf ("\nIntersec: %lf,%lf,%lf\n",lla_i[0],lla_i[1],lla_i[2]);

	//绘制螺旋线
	double alpha = 0;
	double beta = 0;
	int output = 0;
	while (alpha <30)
	{
		//beta为0时的HPR光矢
		double P[3] = {cos(alpha * deg2rad),sin(alpha * deg2rad),0};
		double Hz[3] = {1,0,0},Q[3] = {0,0,0};
		//旋转
		vec_rotation(P,Hz,beta,Q);
		//HPR-->enu
		double Q_enu[3] = {0,0,0};
		HPR2enu(T_torch,Q,Q_enu);
		//enu-->ecef
		double Q_ecef[3] = {0,0,0};
		enu2ecef(R_torch,Q_enu,Q_ecef);
		//求交点
		c_inter = ellipsoid_intersections(ecef_torch,Q_ecef,earth_H,intersecs_ecef,intersecs_t);
		if (c_inter)
		{
			ecef2lla(intersecs_ecef[0],lla_i);
			if (output++ % 10==0)
				printf ("ID=%d, %lf,%lf,%lf\n",output,lla_i[0],lla_i[1],lla_i[2]);
		}


		alpha += 0.1;
		beta += 10;
	}

}

//求取地表到空中某个位置的距离等值线
void demo_contour_distance()
{
	printf ("\n\n========%s=======\n",__FUNCTION__);
	using namespace CES_GEOCALC;
	//光源的LLA
	const double lla_torch [3] = {35.273, 117.121, 10000};

	std::vector<double> betas;
	for (int i=0;i<360;i+=5)
		betas.push_back(i);

	std::vector<double> lats,lons,Ds;

	contour_distance(lla_torch,123000,100,betas,&lats,&lons,&Ds);

	const size_t res_sz = lats.size();

	for (size_t i=0;i<res_sz;++i)
	{
		printf("%.6lf,%.6lf\n",lats[i],lons[i]);
	}
	printf("%.6lf,%.6lf\n",lats[0],lons[0]);
}


void demo_contour_elevation()
{
	printf ("\n\n========%s=======\n",__FUNCTION__);
	using namespace CES_GEOCALC;
	//光源的LLA
	const double lla_torch [3] = {117.121,35.273, 10000};

	std::vector<double> betas;
	for (int i=0;i<360;i+=5)
		betas.push_back(i);

	std::vector<double> lats,lons,Ds,Az,El;

	contour_elevation(lla_torch,15,100,betas,&lats,&lons,&Ds,&Az,&El,true,1e-3,false,false);

	const size_t res_sz = lats.size();

	for (size_t i=0;i<res_sz;++i)
	{
		printf("%.6lf,%.6lf\n",lats[i],lons[i]);
	}
	printf("%.6lf,%.6lf\n",lats[0],lons[0]);
}

